/* *##% 
 * JAXX Maven plugin
 * Copyright (C) 2008 - 2009 CodeLutin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Lesser Public License for more details.
 *
 * You should have received a copy of the GNU General Lesser Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/lgpl-3.0.html>.
 * ##%*/
package org.nuiton.jaxx.plugin;

import java.util.List;
import jaxx.compiler.CompilerConfiguration;
import jaxx.compiler.JAXXCompiler;
import jaxx.compiler.JAXXEngine;
import jaxx.compiler.beans.BeanInfoUtil;
import jaxx.compiler.CompiledObjectDecorator;
import jaxx.compiler.decorators.HelpRootCompiledObjectDecorator;
import jaxx.compiler.tags.TagManager;
import jaxx.runtime.JAXXContext;

import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;

import org.nuiton.io.FileUpdaterHelper;
import org.nuiton.io.MirroredFileUpdater;

import java.io.File;
import java.io.IOException;
import java.util.Map;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Set;
import jaxx.runtime.swing.help.JAXXHelpBroker;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;

/**
 * Generates some java code from jaxx files.
 *
 * @author chemit
 * @goal generate
 * @phase process-sources
 * @requiresDependencyResolution compile
 * @requiresProject
 */
public class GenerateMojo extends AbstractJaxxMojo implements CompilerConfiguration {

    /**
     * Default includes to use, if none provided
     */
    private static final String[] INCLUDES = {"**\\/*.jaxx"};
    /**
     * Repertoire sources des fichiers jaxx a generer.
     *
     * @parameter expression="${jaxx.src}" default-value="${basedir}/src/main/java"
     */
    protected File src;
    /**
     * Repertoire de destination des fichiers java a generer.
     *
     * @parameter expression="${jaxx.outJava}" default-value="${basedir}/target/generated-sources/java"
     */
    protected File outJava;
    /**
     * pour filter les fichiers a traiter
     *
     * @parameter expression="${jaxx.includes}"
     */
    protected String[] includes;
    /**
     * pour filter les fichiers a ne pas traiter
     *
     * @parameter expression="${jaxx.excludes}"
     */
    protected String[] excludes;
    /**
     * Le compilateur à utiliser (par défaut celui de Swing)
     *
     * @parameter expression="${jaxx.compilerFQN}" default-value="jaxx.compiler.JAXXCompiler"
     */
    protected String compilerFQN;
    /**
     * Le compilateur à utiliser (par défaut celui de Swing)
     *
     * @parameter expression="${jaxx.validatorFQN}" default-value="jaxx.runtime.validator.swing.SwingValidator"
     */
    protected String validatorFQN;
    /**
     * the name of implementation of {@link jaxx.runtime.JAXXContext}
     * to be used on {@link jaxx.runtime.JAXXObject}.
     * <p/>
     * Must not be abstract.
     *
     * @parameter expression="${jaxx.jaxxContextFQN}" default-value="jaxx.runtime.context.DefaultJAXXContext"
     * @required
     */
    protected String jaxxContextFQN;
    /**
     * the FQN of the ui to use for error notification.
     * <p/>
     * If not given, will use the one defined in validator
     *
     * @parameter expression="${jaxx.defaultErrorUIFQN}"
     */
    protected String defaultErrorUIFQN;
    /**
     * the FQN of the ui to use for error notification.
     * <p/>
     * If not given, will use the one defined in validator
     *
     * @parameter expression="${jaxx.defaultDecoratorFQN}" default-value="jaxx.compiler.decorators.DefaultCompiledObjectDecorator"
     *
     * @see jaxx.compiler.CompiledObjectDecorator
     */
    protected String defaultDecoratorFQN;
    /**
     * flag to include in compiler classpath the java sources directories (src and outJava).
     * <p/>
     * By default, false.
     *
     * @parameter expression="${jaxx.addSourcesToClassPath}" default-value="false"
     */
    protected boolean addSourcesToClassPath;
    /**
     * flag to include in compiler classpath the java resources directories (src and outJava).
     * <p/>
     * By default, false.
     *
     * @parameter expression="${jaxx.addResourcesToClassPath}" default-value="false"
     * @since 1.6.0
     */
    protected boolean addResourcesToClassPath;
    /**
     * flag to include in compiler classpath the compile class-path (can only be used in a test phase).
     * <p/>
     * By default, false.
     *
     * @parameter expression="${jaxx.addCompileClassPath}" default-value="false"
     * @since 1.6.0
     */
    protected boolean addCompileClassPath;
    /**
     * flag to include in compiler classpath the project compile classpath.
     * <p/>
     * By default, false.
     *
     * @parameter expression="${jaxx.addProjectClassPath}" default-value="false"
     */
    protected boolean addProjectClassPath;
    /**
     * to force generation of java source for any jaxx files with no timestamp checking.
     * <p/>
     * By default, never force generation.
     *
     * @parameter expression="${jaxx.force}" default-value="false"
     */
    protected boolean force;
    /**
     * A flag to mark themojo to be used in a test phase. This will permits to add generated sources in test compile roots.
     *
     * @parameter expression="${jaxx.testPhase}" default-value="false"
     * @since 1.6.0
     */
    protected boolean testPhase;
    /**
     * to make compiler i18nable, says add the {@link org.nuiton.i18n.I18n#_(String, Object...)} method
     * invocation on {@link jaxx.compiler.I18nHelper#I18N_ATTRIBUTES} attributes.
     *
     * @parameter expression="${jaxx.i18nable}" default-value="true"
     * @see jaxx.compiler.I18nHelper
     */
    protected boolean i18nable;
    /**
     * pour optimizer le code compile ou genere ?
     *
     * @parameter expression="${jaxx.optimize}" default-value="false"
     */
    protected boolean optimize;
    /**
     * flag to add logger to each generated jaxx file.
     * <p/>
     * By default, always add it.
     *
     * @parameter expression="${jaxx.addLogger}" default-value="true"
     */
    protected boolean addLogger;
    /**
     * flag to keep compilers after the generate operation (usefull for tests.
     * <p/>
     * By default, always reset.
     *
     * @parameter expression="${jaxx.resetAfterCompile}" default-value="true"
     */
    protected boolean resetAfterCompile;
    /**
     * extra path to be added in {@link java.beans.Introspector#setBeanInfoSearchPath(java.lang.String[])}.
     * <p/>
     * add beanInfoSearchPath to be registred by {@link BeanInfoUtil#addJaxxBeanInfoPath(java.lang.String...)}
     * <p/>
     * and then will be use by {@link jaxx.compiler.spi.DefaultInitializer#initialize()}.
     * <p/>
     * <p/>
     * This permit to use real beanInfo of imported graphic libraries.
     *
     * @parameter expression="${jaxx.beanInfoSearchPath}"
     */
    protected String[] beanInfoSearchPath;
    /**
     * list of fqn of class toimport for all generated jaxx files
     *
     * @parameter expression="${jaxx.extraImportList}"
     */
    protected String extraImportList;
    /**
     * a flag to use UIManager to retreave icons.
     *
     * @parameter expression="${jaxx.useUIManagerForIcon}" default-value="false"
     */
    protected boolean useUIManagerForIcon;
    /**
     * flag to activate profile mode.
     * <p/>
     * By default, not active.
     *
     * @parameter expression="${jaxx.profile}" default-value="false"
     */
    protected boolean profile;
    /**
     * flag to activate help generation process.
     * <p/>
     * By default, not active.
     *
     * @parameter expression="${jaxx.generateHelp}" default-value="false"
     *
     * @since 1.3
     */
    protected boolean generateHelp;
    /**
     * the FQN of help broker
     * <p/>
     * By default, use the JAXX implementation {@link JAXXHelpBroker}.
     *
     * @parameter expression="${jaxx.helpBrokerFQN}" default-value="jaxx.runtime.swing.help.JAXXHelpBroker"
     *
     * @since 1.3
     */
    protected String helpBrokerFQN;
    /**
     * detected jaxx files in {@link #init()} method
     */
    protected String[] files;
    /**
     * file updater used to detect jaxx files.
     *
     * <b>Note:</b> if {@link #verbose} flag is on, will ne be used
     */
    protected MirroredFileUpdater updater;
    /**
     * 
     */
    private Class<?> defaultErrorUIClass;
    /**
     *
     */
    private Class<?> validatorClass;
    /**
     *
     */
    private Class<? extends CompiledObjectDecorator> defaultDecoratorClass;
    /**
     *
     */
    private Class<? extends JAXXContext> jaxxContextClass;
    /**
     *
     */
    private Class<? extends JAXXCompiler> compilerClass;
    /**
     *
     */
    private String[] extraImports;
    /**
     *
     */
    private boolean nofiles;
    /**
     *
     */
    protected ClassLoader cl;
    /**
     * JAXX engine
     */
    private JAXXEngine engine;

    @SuppressWarnings("unchecked")
    @Override
    public boolean init() throws Exception {

        if (generateHelp) {
            // check there is some bundle
            if (getHelpIdsStore() == null) {
                throw new MojoFailureException("you must set the 'helpIdStore' property.");
            }
        }

        fixCompileSourceRoots();

        if (includes == null || includes.length == 0) {
            // use default includes
            includes = INCLUDES;
        }
        updater = FileUpdaterHelper.newJaxxFileUpdater(src, outJava);

        Map<File, String[]> result = new HashMap<File, String[]>();
        getFilesToTreateForRoots(includes, excludes, Arrays.asList(src.getAbsolutePath()), result, force ? null : updater);

        this.files = result.get(src);

        nofiles = files == null || files.length == 0;
        if (nofiles) {
            return true;
        }

        cl = initClassLoader(project,
                src,
                addSourcesToClassPath,
                testPhase,
                addResourcesToClassPath,
                addCompileClassPath,
                addProjectClassPath);

        Thread.currentThread().setContextClassLoader(cl);

        compilerClass = (Class<? extends JAXXCompiler>) Class.forName(compilerFQN, false, cl);
        defaultDecoratorClass = (Class<? extends CompiledObjectDecorator>) Class.forName(defaultDecoratorFQN, false, cl);
        jaxxContextClass = (Class<? extends JAXXContext>) Class.forName(jaxxContextFQN, false, cl);
        if (!JAXXContext.class.isAssignableFrom(jaxxContextClass)) {
            throw new MojoExecutionException("jaxxContextFQN must be an implementation of " + JAXXContext.class + " but was : " + jaxxContextClass);
        }
        validatorClass = Class.forName(validatorFQN, false, cl);

        if (defaultErrorUIFQN != null && !defaultErrorUIFQN.trim().isEmpty()) {
            defaultErrorUIClass = Class.forName(defaultErrorUIFQN, false, cl);
        }

        if (beanInfoSearchPath != null && beanInfoSearchPath.length > 0) {
            // register extra path
            BeanInfoUtil.addJaxxBeanInfoPath(beanInfoSearchPath);
        }

        if (!getTargetDirectory().exists()) {
            getTargetDirectory().mkdirs();
        }

        // compute extra imports (can not use java classes since some of
        // imports can not be still compiled)
        if (extraImportList != null && !extraImportList.isEmpty()) {
            String[] imports = extraImportList.split(",");
            int i = 0;
            for (String importS : imports) {
                imports[i++] = importS.trim();
            }
            if (verbose) {
                getLog().info("extra imports " + java.util.Arrays.toString(imports));
            }
            extraImports = imports;
        }

        if (verbose) {
            getLog().info(toString());
            getLog().info("includes : " + Arrays.toString(includes));
            for (String file : files) {
                getLog().info("will parse " + file);
            }
        }
        return true;
    }

    @Override
    public void doAction() throws MojoExecutionException {

        if (nofiles) {
            getLog().info("No files to treate.");
            return;
        }

        getLog().info("Detects " + files.length + " modified jaxx file(s). ");

        try {

            // force compiler init from here, not in a static block
            TagManager.reset(verbose);

            engine = JAXXEngine.newLaunchor(src, files, this);
            int nbFiles = engine.run();
            report(engine);
            if (nbFiles == -1) {
                throw new MojoExecutionException("Aborting due to errors reported by jaxxc");
            }
            getLog().info("Generated " + nbFiles + " file(s). ");

            if (generateHelp) {
                // generate help
                generateHelp();
            }

        } catch (MojoExecutionException e) {
            getLog().error(e);
            throw e;
        } catch (Exception e) {
            //getLog().error(e);
            Throwable e2 = e;
            while (e2.getCause() != null) {
                e2 = e.getCause();
            }
            getLog().error(e2);

            throw new MojoExecutionException(e2.getMessage(), e2);
        }
    }

    @Override
    public File getTargetDirectory() {
        return outJava;
    }

    @Override
    public void setTargetDirectory(File targetDirectory) {
        this.outJava = targetDirectory;
    }

    @Override
    public boolean getOptimize() {
        return optimize;
    }

    @Override
    public boolean isI18nable() {
        return i18nable;
    }

    @Override
    public boolean isUseUIManagerForIcon() {
        return useUIManagerForIcon;
    }

    @Override
    public boolean isAddLogger() {
        return addLogger;
    }

    @Override
    public Class<? extends JAXXContext> getJaxxContextClass() {
        return jaxxContextClass;
    }

    @Override
    public String[] getExtraImports() {
        return extraImports;
    }

    @Override
    public boolean isResetAfterCompile() {
        return resetAfterCompile;
    }

    @Override
    public boolean isOptimize() {
        return optimize;
    }

    @Override
    public Class<?> getDefaultErrorUI() {
        return defaultErrorUIClass;
    }

    @Override
    public ClassLoader getClassLoader() {
        return cl;
    }

    @Override
    public Class<? extends JAXXCompiler> getCompilerClass() {
        return compilerClass;
    }

    @Override
    public Class<? extends CompiledObjectDecorator> getDefaultDecoratorClass() {
        return defaultDecoratorClass;
    }

    @Override
    public boolean isProfile() {
        return profile;
    }

    @Override
    public boolean isGenerateHelp() {
        return generateHelp;
    }

    @Override
    public String getHelpBrokerFQN() {
        return helpBrokerFQN;
    }

    @Override
    public Class<?> getValidatorClass() {
        return validatorClass;
    }

    public JAXXEngine getEngine() {
        return engine;
    }

    @Override
    public String toString() {
        return ToStringBuilder.reflectionToString(this, ToStringStyle.MULTI_LINE_STYLE);
    }

    protected void fixCompileSourceRoots() {
        if (project == null) {
            // no project defined, can not fix anything
            // this case could appear if we wanted to do some tests of the plugin
            return;
        }

        if (testPhase) {
            addTestCompileSourceRoots(getTargetDirectory());
        } else {
            addCompileSourceRoots(getTargetDirectory());
        }
    }

    protected void generateHelp() throws IOException {
        Set<String> helpIds = HelpRootCompiledObjectDecorator.getHelpIds();
        if (helpIds.isEmpty()) {
            if (verbose) {
                // no ids detected in this compilation round
                getLog().info("no helpIds detected.");
            }
            return;
        }
        File idsStore = getHelpIdsStore();

        if (!idsStore.getParentFile().exists()) {
            idsStore.getParentFile().mkdirs();
        }

        StringBuilder buffer = new StringBuilder();

        for (String helpId : helpIds) {
            buffer.append(removeQuote(helpId)).append('\n');
        }
        writeFile(idsStore, buffer.toString(), encoding);

        getLog().info("helpIdStore generated in " + idsStore);

        helpIds.clear();
    }

    protected String removeQuote(String txt) {
        if (txt.startsWith("\"")) {
            txt = txt.substring(1);
        }
        if (txt.endsWith("\"")) {
            txt = txt.substring(0, txt.length() - 1);
        }
        return txt;
    }

    protected void report(JAXXEngine engine) {
        List<String> warnings = engine.getWarnings();
        if (!warnings.isEmpty()) {
            StringBuilder buffer = new StringBuilder("JAXX detects " + (warnings.size() == 1 ? "1 warning" : warnings.size() + " warnings"));
            buffer.append(" :");
            for (String s : warnings) {
                buffer.append("\n").append(s);
            }
            getLog().warn(buffer.toString());
        }
        List<String> errors = engine.getErrors();
        if (!errors.isEmpty()) {
            StringBuilder buffer = new StringBuilder("JAXX detects " + (errors.size() == 1 ? "1 error" : errors.size() + " errors"));
            buffer.append(" :");
            for (String s : errors) {
                buffer.append("\n").append(s);
            }
            getLog().error(buffer.toString());
        }
    }
}
